#include <stdio.h>
#include <stdint.h>
#include <stdarg.h>
#include <limits.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <vector>
#include <mpi.h>

#include "solver.h"
#ifdef MKL
#include "mkl.h"
#endif

template <class T> static inline void swap(T* a, int x, int y) { T t=a[x]; a[x]=a[y]; a[y]=t; }
#ifndef TIMER
#define TIMER
#define NS_PER_SEC 1000000000
double wall_time_diff(int64_t ed, int64_t st)
{
	return (double)(ed-st)/(double)NS_PER_SEC;
}

int64_t wall_clock_ns()
{
#if __unix__
	struct timespec tspec;
	clock_gettime(CLOCK_MONOTONIC, &tspec);
	return tspec.tv_sec*NS_PER_SEC + tspec.tv_nsec;
#else
#if __MACH__
	return 0;
#else
	struct timeval tv;
	gettimeofday( &tv, NULL );
	return tv.tv_sec*NS_PER_SEC + tv.tv_usec*1000;
#endif
#endif
}
#endif

#ifndef min
template <class T> static inline T min(T x,T y) { return (x<y)?x:y; }
#endif

#ifndef max
template <class T> static inline T max(T x,T y) { return (x>y)?x:y; }
#endif

#define MEPS 1e-16L
#define PIVOTING
#define DEBUG_LDL

// solve Ax=b , in which A[i*n+j] = Aij and we only solve submatrix A[m:m]
double solve_system(const double *_A, const double *_b, int n, int m, double *x)
{
	long double *A = new long double[n * n];
	for(int i=0; i<n*n; i++)
		A[i] = _A[i];
	long double *b = new long double[n];
	for(int i=0; i<n; i++)
		b[i] = _b[i];

#ifdef PIVOTING
	int *p = new int[m];
	for(int i=0; i<m; i++)
		p[i] = i;
#endif

	// decomp
	// A -> D0
	//      L00 D1
	//      L10 L11 D2
	for(int j=0; j<m; j++){
#ifdef PIVOTING
		int pj = j;
		long double pj_val = fabsl(A[j*n+j]);
		// find piv
		for(int i=j+1; i<m; i++){
			long double pi_val = fabsl(A[i*n+i]);
			if(pj_val < pi_val){
				pj_val = pi_val;
				pj = i;
			}
		}
		// swap rows and columns of j and pj
		for(int i=0; i<m; i++)
			swap(A, j*n+i, pj*n+i);
		for(int i=0; i<m; i++)
			swap(A, i*n+j, i*n+pj);
		swap(b, j, pj);
		swap(p, j, pj);
#endif

		// loop invariant: Aik = Lik, Akk = Dk with k<j
		long double Dj = A[j*n+j];
		for(int k=0; k<j; k++){
			Dj -= A[j*n+k]*A[j*n+k] * A[k*n+k];
		}
		A[j*n+j] = Dj;
		//assert(fabs(Dj) >= MEPS);
		if(fabsl(Dj) < MEPS){
			for(int i=j+1; i<m; i++)
				A[i*n+j] = 0;
			continue; // xj should be 0
		}

		for(int i=j+1; i<m; i++){
			long double Lij = A[i*n+j];
			for(int k=0; k<j; k++){
				Lij -= A[i*n+k] * A[j*n+k] * A[k*n+k];
			}
			Lij /= Dj;
			A[i*n+j] = Lij;
		}
	}
#if 0
	printf("L\n");
	for(int i=0; i<n; i++){
		for(int j=0; j<=i; j++)
			printf("%+.3Le ", A[i*n+j]);
		printf("\n");
	}
#endif
	// invariant: A[i*n+j] = Lij, forall i>j
	//            A[i*n+i] = Dii, forall i

	// LD y = b
	// di yi =  bi - \sum_{j<i} yj Lij
	// place y in b
	for(int i=0; i<m; i++){
		// loop invariant: bk = yk forall k<i
		long double yi = b[i];
		for(int k=0; k<i; k++)
			yi -= A[i*n+k] * b[k];
		b[i] = yi;
	}
	for(int i=0; i<m; i++){
		long double Di = A[i*n+i];
		if(fabsl(Di) < MEPS)
			b[i] = 0;
		else
			b[i] /= Di;
	}
#if 0
	printf("y\n");
	for(int i=0; i<n; i++){
		printf("%+.3e\n", b[i]);
	}
#endif
	// L'x = y
	// xi = yi - \sum{j>i} xj Lij
	for(int i=m-1; i>=0; i--){
		// loop invariant: xk ready forall k>i
		long double xi = b[i];
#ifdef PIVOTING
		for(int k=m-1; k>i; k--)
			xi -= A[k*n+i] * x[p[k]];
		x[p[i]] = (double)xi;
#else
		for(int k=m-1; k>i; k--)
			xi -= A[k*n+i] * x[k];
		x[i] = (double)xi;
#endif
	}
#if 0
	printf("\n");
	for(int i=0; i<n; i++)
		printf("d %.3e\n", x[i]);

	printf("A\n");
	for(int i=0; i<n; i++){
		for(int j=0; j<n; j++)
			printf("%+.3e ", Acopy[i*n+j]);
		printf("\n");
	}
	printf("b\n");
	for(int i=0; i<n; i++){
		printf("%+.3e\n", bcopy[i]);
	}
#endif

#if 1
	double err = 0;
	for(int i=0; i<m; i++){
		double Ax_i = 0;
		for(int j=0; j<m; j++)
			Ax_i += _A[i*n+j] * x[j];
		err += fabs(Ax_i-_b[i]);
	}
	double max_pv = 0;
	for(int i=0; i<m; i++)
		if(max_pv < _A[i*n+i])
			max_pv = (double)_A[i*n+i];
	if(max_pv != 0)
		err /= max_pv;
	//printf("err=%e\n", err);
	//assert(err < 1e-5);
#endif
	delete[] A;
	delete[] b;
#ifdef PIVOTING
	delete[] p;
#endif
	return err;
}

static int mpi_get_rank()
{
	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	return rank;	
}

static int mpi_get_size()
{
	int size;
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	return size;	
}

template<typename T>
void static mpi_allreduce(T *buf, const int count, MPI_Datatype type, MPI_Op op)
{
	size_t count_t = size_t(count);
	std::vector<T> buf_reduced(count_t);
	MPI_Allreduce(buf, buf_reduced.data(), count, type, op, MPI_COMM_WORLD);
	memcpy(buf, buf_reduced.data(), sizeof(T) * count_t);
}

#ifdef MKL
#include "mkl.h"
#define dnrm2_ dnrm2
#define ddot_ ddot
#define daxpy_ daxpy
#define dscal_ dscal
#define dgemv_ dgemv
#define dsyrk_ dsyrk
#define dspmv_ dspmv
#else
#ifdef __cplusplus
extern "C" {
#endif

extern double dnrm2_(int *, double *, int *);
extern double ddot_(int *, double *, int *, double *, int *);
extern int daxpy_(int *, double *, double *, int *, double *, int *);
extern int dscal_(int *, double *, double *, int *);
extern int dgemv_(char *, int *, int *, double *, double *, int *, double *, int *, double *, double *, int *);
extern int dsyrk_(char *, char *, int *, int *, double *, double *, int *, double *, double *, int *);
extern int dspmv_(char *, int *, double *, double *, double *, int *, double *, double *, int *);

#ifdef __cplusplus
}
#endif
#endif

static void default_print(const char *buf)
{
	fputs(buf,stdout);
	fflush(stdout);
}

static double uTMv(int n, double *u, double *M, double *v)
{
	const int m = n-4;
	double res = 0;
	int i;
	for (i=0; i<m; i+=5)
		res += u[i]*M[i]*v[i]+u[i+1]*M[i+1]*v[i+1]+u[i+2]*M[i+2]*v[i+2]+
			u[i+3]*M[i+3]*v[i+3]+u[i+4]*M[i+4]*v[i+4];
	for (; i<n; i++)
		res += u[i]*M[i]*v[i];
	return res;
}

void SOLVER::info(const char *fmt,...)
{
	if(mpi_get_rank()!=0)
		return;
	char buf[BUFSIZ];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buf,fmt,ap);
	va_end(ap);
	(*solver_print_string)(buf);
}

void SOLVER::set_print_string(void (*print_string) (const char *buf))
{
	solver_print_string = print_string;
}

SOLVER::SOLVER(const function *fun_obj, double eps, int max_iter)
{
	this->fun_obj = const_cast<function *>(fun_obj);
	this->eps = eps;
	this->max_iter = max_iter;
	solver_print_string = default_print;
	int nr_node = mpi_get_size();
	int rank = mpi_get_rank();
	int n = this->fun_obj->get_nr_variable();
	int shift = int(ceil(double(n) / double(nr_node)));
	start = shift * rank;
	length = min(max(n - start, 0), shift);
	if (length == 0)
		start = 0;
#ifdef MKL
	mkl_set_num_threads(1);
#endif
}

SOLVER::~SOLVER()
{
}


MLBFGS::MLBFGS(const function *fun_obj, int l, double eps, int m, double eta, int max_iter): SOLVER(fun_obj, eps, max_iter)
{
	this->eta = eta;
	this->M = m;
	this->l = l;
	Plength = 2 * M + 2;
	this->n = this->fun_obj->get_nr_variable();
	int nr_node = mpi_get_size();
	int shift = int(ceil(double(n) / double(nr_node)));
	recv_count = new int[nr_node];
	displace = new int[nr_node];
	recv_buffer = new double[n];
	int counter = 0;
	for (int i=0;i<nr_node;i++)
	{
		recv_count[i] = shift;
		displace[i] = counter;
		counter += shift;
		if (counter >= n)
		{
			counter = 0;
			shift = 0;
		}
		else if (counter + shift > n)
			shift = n - counter;
	}
}

MLBFGS::~MLBFGS()
{
	delete[] recv_buffer;
	delete[] recv_count;
	delete[] displace;
}

void MLBFGS::solve(double *w)
{
	sqrtD_Px = new double[Plength * l];
	int i;
	double f;
	double one = 1.0;
	double minusone = -1.0;
	double zero = 0.0;
	int iter = 0, inc = 1;
	int DynamicM = 0;
	char N[] = "N";
	char U[] = "U";
	char T[] = "T";
	double gnorm;
	double step_size = 0;
	double origDirDeriv = 0;
	double err = 0;
	double dTd;
	double gnorm1;
	int previous_M = 0;
	int index_length = 0;
	int *index = NULL;

	wcounter = 0;
	gcounter = 1;
	
	double* linear = new double[Plength];
	double* dTdtmp = new double[Plength];
	double *sqrt_C_times_D = new double[l];
	double *inner_product_matrix = new double[Plength * Plength];//symmetric matrix, so only need to store the upper triangular part
	double *step = new double[length];
	double *step_x = new double[l];
	double *delta = new double[Plength];
	double *P = new double[Plength * length];
	double *Px = new double[Plength * l];
	// The order is w_i, g_i, w_{i+1}, g_{i+1} ...,
	// Each column represents a possible direction to update, stored in a 1d array but treated as an n * length matrix as in blas
	double *localw = P;
	double *wx = Px;
	double *g = P + gcounter * length;
	double *gx = Px + gcounter * l;
	double *system = new double[Plength * Plength];
	int sparse;
	double *global_g = new double[n];

	tmp_length = 2;
	memset(w, 0, sizeof(double) * size_t(n));
	memcpy(localw, w + start, sizeof(double) * size_t(length));
	memset(wx, 0, sizeof(double) * size_t(l));
	int64_t time_st = wall_clock_ns();

	f = fun_obj->fun(w);
	fun_obj->grad(w, global_g);
	fun_obj->Xv(global_g, gx);
	memcpy(g, global_g + start, sizeof(double) * size_t(length));
	gnorm = ddot_(&length, g, &inc, g, &inc);
	mpi_allreduce(&gnorm, 1, MPI_DOUBLE, MPI_SUM);
	gnorm = sqrt(gnorm);
	gnorm1 = gnorm;
	double tmp[2] = {0, 0};
	dgemv_(T, &length, &tmp_length, &one, P, &length, localw, &inc, &zero, tmp, &inc);
	mpi_allreduce(tmp, 2, MPI_DOUBLE, MPI_SUM);
	inner_product_matrix[inner_product_entry(wcounter, wcounter)] = tmp[wcounter];
	inner_product_matrix[inner_product_entry(wcounter, gcounter)] = tmp[gcounter];
	inner_product_matrix[inner_product_entry(gcounter, gcounter)] = gnorm * gnorm;

	int64_t iter_ed = wall_clock_ns();
	double tt = wall_time_diff(iter_ed,time_st);
	time_st = wall_clock_ns();

	while (iter < max_iter)
	{
		gnorm = sqrt(inner_product_matrix[inner_product_entry(gcounter, gcounter)]);

		int64_t timer_ed = wall_clock_ns();
		tt += wall_time_diff(timer_ed, time_st);
		time_st = wall_clock_ns();
		info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
		if (gnorm <= eps*gnorm1)
			break;
		sparse = (iter <= M);
		tmp_length -= (DynamicM + 1) * sparse;

		fun_obj->sqrt_Hessian_coeff(sqrt_C_times_D, &index, &index_length);
		if (index_length == 0)
			compute_system(inner_product_matrix, Px, sqrt_C_times_D, system, sparse);
		else
			compute_system(inner_product_matrix, Px, sqrt_C_times_D, system, sparse, index, index_length);
		

		for (i=0;i<tmp_length;i++)
			for (int j=0;j<tmp_length;j++)
				info("system[%d][%d] = %g\n",i,j,system[i * tmp_length + j]);

		if (sparse)
			for (i=0;i<tmp_length;i++)
				linear[i] = -inner_product_matrix[inner_product_entry(2 * i + 1, gcounter)];
		else
			for (i=0;i<tmp_length;i++)
				linear[i] = -inner_product_matrix[inner_product_entry(i, gcounter)];

		for (i=0;i<tmp_length;i++)
			info("linear[%d] = %g\n",i,linear[i]);



		err = solve_system(system, linear, tmp_length, tmp_length, delta);

		if (sparse)
		{
			for (i=tmp_length;i>0;i--)
			{
				delta[2 * i - 1] = delta[i - 1];
				delta[2 * i - 2] = 0;
			}
			tmp_length = 2 * DynamicM + 2;
		}
		origDirDeriv = 0;
		memset(dTdtmp, 0, sizeof(double) * Plength);
		dspmv_(U, &tmp_length, &one, inner_product_matrix, delta, &inc, &zero, dTdtmp, &inc);
		dTd = ddot_(&tmp_length, delta, &inc, dTdtmp, &inc);
		origDirDeriv = dTdtmp[gcounter];
		double ws = dTdtmp[wcounter];

		if (origDirDeriv > 0)
		{
			origDirDeriv = -origDirDeriv;
			ws = -ws;
			dscal_(&tmp_length, &minusone, delta, &inc);
		}
		if (fabs(origDirDeriv) < inner_product_matrix[inner_product_entry(gcounter, gcounter)] * 1e-12)
		{
			memset(delta, 0, sizeof(double) * Plength);
			delta[gcounter] = -1;
			dTd = inner_product_matrix[inner_product_entry(gcounter, gcounter)];
			origDirDeriv = -dTd;
			ws  = -inner_product_matrix[inner_product_entry(wcounter, gcounter)];
			memset(step, 0, sizeof(double) * length);
			memset(step_x, 0, sizeof(double) * l);
			daxpy_(&length, &minusone, g, &inc, step, &inc);
			daxpy_(&l, &minusone, gx, &inc, step_x, &inc);
		}
		else
		{
			dgemv_(N, &length, &tmp_length, &one, P, &length, delta, &inc, &zero, step, &inc);
			dgemv_(N, &l, &tmp_length, &one, Px, &l, delta, &inc, &zero, step_x, &inc);
		}
		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		MPI_Allgatherv(step, length, MPI_DOUBLE, recv_buffer, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);
		fun_obj->Xv(recv_buffer, step_x);
		time_st = wall_clock_ns();

		double oldf = f;
		step_size = fun_obj->line_search(w, step, origDirDeriv, dTd, ws, &f, eta, 1.0, step_x);

		if (step_size == 0)
		{
			info("WARNING: step size too small\n");
			break;
		}
		double actred = oldf - f;
		if (fabs(actred) < 0)
		{
			info("WARNING: actred < 0\n");
			break;
		}
		if (fabs(actred) <= 1.0e-20*fabs(f))
		{
			info("WARNING: actred too small\n");
			break;
		}

		iter++;
		previous_M = DynamicM;
		DynamicM = min(iter, M);
		tmp_length = 2 * DynamicM + 2;

		wcounter = (wcounter + 2) % Plength;
		gcounter = wcounter + 1;
		if (step_size == 1)
		{
			double *tmp_ptr = P + wcounter * length;
			for (i=0;i<length;i++)
				tmp_ptr[i] = localw[i] + step[i];
			localw = P + wcounter * length;
		}
		else
		{
			memcpy(P + wcounter * length, localw, sizeof(double) * length);
			localw = P + wcounter * length;
			daxpy_(&length, &step_size, step, &inc, localw, &inc);
		}
		wx = Px + wcounter * l;

		g = P + gcounter * length;
		gx = Px + gcounter * l;
		memcpy(w + start, localw, sizeof(double) * length);
		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		if (iter % 10 == 0 && iter != 0)
		{
			MPI_Allgatherv(localw, length, MPI_DOUBLE, recv_buffer, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);
			memcpy(w, recv_buffer, sizeof(double) * size_t(n));
			f = fun_obj->fun(w);
		}
		fun_obj->get_wx(wx);
		time_st = wall_clock_ns();
		fun_obj->grad(w, global_g);

		fun_obj->Xv(global_g, gx);
		memcpy(g, global_g + start, sizeof(double) * size_t(length));

		update_inner_products(inner_product_matrix, DynamicM, previous_M, P, delta, step_size);
	}
	MPI_Allgatherv(localw, length, MPI_DOUBLE, recv_buffer, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);
	memcpy(w, recv_buffer, sizeof(double) * size_t(n));

	delete[] dTdtmp;
	delete[] delta;
	delete[] system;
	delete[] Px;
	delete[] P;
	delete[] step;
	delete[] step_x;
	delete[] sqrt_C_times_D;
	delete[] linear;
	delete[] inner_product_matrix;
	delete[] sqrtD_Px;
	delete[] global_g;
}

void MLBFGS::compute_system(double* inner_product_matrix, double* Px, double* sqrt_C_times_D, double* system, int sparse)
{
	double one = 1.0;
	double zero = 0;
	int i,j;
	char U[] = "U";
	char T[] = "T";
	double *tmp = new double[tmp_length * (tmp_length + 1) / 2];
	if (sparse)
		for (i=0;i<tmp_length;i++)
		{
			int counteri = i * l;
			int tmp_counter = (2 * i + 1) * l;
			for (j=0;j<l;j++)
				sqrtD_Px[counteri + j] = sqrt_C_times_D[j] * Px[tmp_counter + j];
		}
	else
		for (i=0;i<tmp_length;i++)
		{
			int counteri = i * l;
			for (j=0;j<l;j++)
				sqrtD_Px[counteri + j] = sqrt_C_times_D[j] * Px[counteri + j];
		}
	dsyrk_(U, T, &tmp_length, &l, &one, sqrtD_Px, &l, &zero, system, &tmp_length);

	int counter = 0;
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			tmp[counter++] = system[i * tmp_length + j];
	mpi_allreduce(tmp, counter, MPI_DOUBLE, MPI_SUM);
	counter = 0;
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[i * tmp_length + j] = tmp[counter++];
	if (sparse)
		for (i=0;i<tmp_length;i++)
			for (j=0;j<=i;j++)
				system[i * tmp_length + j] += inner_product_matrix[inner_product_entry(2 * i + 1, 2 * j + 1)];
	else
		for (i=0;i<tmp_length;i++)
			for (j=0;j<=i;j++)
				system[i * tmp_length + j] += inner_product_matrix[inner_product_entry(i,j)];

	for (i=0;i<tmp_length;i++)
		for (j=i+1;j<tmp_length;j++)
			system[i * tmp_length + j] = system[j * tmp_length + i];
	delete[] tmp;
}

void MLBFGS::compute_system(double* inner_product_matrix, double* Px, double* sqrt_C_times_D, double* system, int sparse, int *index, int index_length)
{
	double one = 1.0;
	double zero = 0;
	int i,j;
	double *tmp = new double[tmp_length * (tmp_length + 1) / 2];
	char U[] = "U";
	char T[] = "T";
	if (sparse)
		for (i=0;i<tmp_length;i++)
		{
			int counteri = i * index_length;
			int tmp_counter = (2 * i + 1) * l;
			for (j=0;j<index_length;j++)
				sqrtD_Px[counteri + j] = sqrt_C_times_D[j] * Px[tmp_counter + index[j]];
		}
	else
		for (i=0;i<tmp_length;i++)
		{
			int counteri = i * index_length;
			int tmp_counter = i * l;
			for (j=0;j<index_length;j++)
				sqrtD_Px[counteri + j] = sqrt_C_times_D[j] * Px[tmp_counter + index[j]];
		}
	dsyrk_(U, T, &tmp_length, &index_length, &one, sqrtD_Px, &index_length, &zero, system, &tmp_length);

	int counter = 0;
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			tmp[counter++] = system[i * tmp_length + j];
	mpi_allreduce(tmp, counter, MPI_DOUBLE, MPI_SUM);
	counter = 0;
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[i * tmp_length + j] = tmp[counter++];

	if (sparse)
		for (i=0;i<tmp_length;i++)
			for (j=0;j<=i;j++)
				system[i * tmp_length + j] += inner_product_matrix[inner_product_entry(2 * i + 1, 2 * j + 1)];
	else
		for (i=0;i<tmp_length;i++)
			for (j=0;j<=i;j++)
				system[i * tmp_length + j] += inner_product_matrix[inner_product_entry(i,j)];

	for (i=0;i<tmp_length;i++)
		for (j=i+1;j<tmp_length;j++)
			system[i * tmp_length + j] = system[j * tmp_length + i];
	delete[] tmp;
}

inline int MLBFGS::inner_product_entry(int i, int j)
{
	int index1, index2;
	index1 = (i > j) * i + (i <= j) * j;
	index2 = i + j - index1;
	return index1 * (index1 + 1) / 2 + index2;
}

void MLBFGS::update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* delta, double step_size)
{
	int i;
	int prev_length = 2 * previous_M + 2;
	int inc = 1;
	double *PTg = new double[tmp_length];
	double *PTw = new double[tmp_length];
	double one = 1.0;
	double zero = 0;
	char U[] = "U";
	char T[] = "T";
	dgemv_(T, &length, &tmp_length, &one, P, &length, P + gcounter * length, &inc, &zero, PTg, &inc);
	mpi_allreduce(PTg, tmp_length, MPI_DOUBLE, MPI_SUM);

	int prev_w = wcounter - 2;
	if (wcounter == 0)
		prev_w = 2 * DynamicM;

	if (step_size != 1)
		dscal_(&prev_length, &step_size, delta, &inc);
	delta[prev_w] += 1;

	dspmv_(U, &prev_length, &one, inner_product_matrix, delta, &inc, &zero, PTw, &inc);
	PTw[wcounter] = ddot_(&prev_length, PTw, &inc, delta, &inc);

	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, wcounter)] = PTw[i];
	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, gcounter)] = PTg[i];


	delete[] PTg;
	delete[] PTw;
}

MLBFGS_DIAG::MLBFGS_DIAG(const function *fun_obj, int l, double eps, int m, double eta, int max_iter): MLBFGS(fun_obj, l, eps, m, eta, max_iter)
{
	this->M = m;
	Plength = 3 * M + 3;
	this->n = this->fun_obj->get_nr_variable();
}

MLBFGS_DIAG::~MLBFGS_DIAG()
{
}

void MLBFGS_DIAG::solve(double *w)
{
	sqrtD_Px = new double[Plength * l];
	int i;
	double f;
	double one = 1.0;
	double minusone = -1.0;
	double zero = 0.0;
	int iter = 0, inc = 1;
	int DynamicM = 0;
	char N[] = "N";
	char U[] = "U";
	char T[] = "T";
	double gnorm;
	double step_size = 0;
	double origDirDeriv = 0;
	double err = 0;
	double dTd;
	double gnorm1;
	double tt = 0;
	int previous_M = 0;
	int index_length = 0;
	int *index = NULL;

	wcounter = 0;
	gcounter = 1;
	dcounter = 2;
	
	double* linear = new double[Plength];
	double* dTdtmp = new double[Plength];
	double *sqrt_C_times_D = new double[l];
	double *inner_product_matrix = new double[Plength * Plength];//symmetric matrix, so only need to store the upper triangular part
	double *step = new double[length];
	double *step_x = new double[l];
	double *delta = new double[Plength];
	double *P = new double[Plength * length];
	double *Px = new double[Plength * l];
	// The order is w_i, g_i, w_{i+1}, g_{i+1} ...,
	// Each column represents a possible direction to update, stored in a 1d array but treated as an n * length matrix as in blas
	double *localw = P;
	double *wx = Px;
	double *g = P + gcounter * length;
	double *gx = Px + gcounter * l;
	double *D_inverse_g = P + dcounter * length;
	double *D_inverse_g_x = Px + dcounter * l;
	double *global_g = new double[2 * n];
	double *D = global_g + n;
	double *system = new double[Plength * Plength];
	int sparse;

	tmp_length = 3;
	memset(w, 0, sizeof(double) * size_t(n));
	memcpy(localw, w + start, sizeof(double) * size_t(length));

	f = fun_obj->fun(w);
	fun_obj->grad(w, global_g);
	memcpy(g, global_g + start, sizeof(double) * size_t(length));
	gnorm = ddot_(&length, g, &inc, g, &inc);
	mpi_allreduce(&gnorm, 1, MPI_DOUBLE, MPI_SUM);
	gnorm = sqrt(gnorm);
	gnorm1 = gnorm;
	fun_obj->get_diagH(D);
	for (i=0;i<n;i++)
		D[i] = global_g[i] / D[i];
	fun_obj->Xv(global_g, gx, 2);

	memcpy(D_inverse_g, D + start, sizeof(double) * size_t(length));
	

	double tmp[5];
	int two = 2;

	dgemv_(T, &length, &tmp_length, &one, P, &length, P + dcounter * length, &inc, &zero, tmp, &inc);
	dgemv_(T, &length, &two, &one, P, &length, P + wcounter * length, &inc, &zero, tmp + 3, &inc);
	mpi_allreduce(tmp, 5, MPI_DOUBLE, MPI_SUM);
	inner_product_matrix[inner_product_entry(dcounter, wcounter)] = tmp[wcounter];
	inner_product_matrix[inner_product_entry(dcounter, gcounter)] = tmp[gcounter];
	inner_product_matrix[inner_product_entry(dcounter, dcounter)] = tmp[dcounter];
	inner_product_matrix[inner_product_entry(wcounter, wcounter)] = tmp[wcounter + 3];
	inner_product_matrix[inner_product_entry(wcounter, gcounter)] = tmp[gcounter + 3];
	inner_product_matrix[inner_product_entry(gcounter, gcounter)] = gnorm * gnorm;


	int64_t time_st = wall_clock_ns();
	while (iter < max_iter)
	{
		gnorm = sqrt(inner_product_matrix[inner_product_entry(gcounter, gcounter)]);

		int64_t timer_ed = wall_clock_ns();
		tt += wall_time_diff(timer_ed, time_st);
		time_st = wall_clock_ns();
		info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
		if (gnorm <= eps*gnorm1)
			break;

		sparse = (iter <= M);
		tmp_length -= (DynamicM + 1) * sparse;

		fun_obj->sqrt_Hessian_coeff(sqrt_C_times_D, &index, &index_length);
		if (index_length == 0)
			compute_system(inner_product_matrix, Px, sqrt_C_times_D, system, sparse);
		else
			compute_system(inner_product_matrix, Px, sqrt_C_times_D, system, sparse, index, index_length);

		if (sparse)
			for (i=0;i<tmp_length / 2;i++)
			{
				linear[2 * i] = -inner_product_matrix[inner_product_entry(3 * i + 1, gcounter)];
				linear[2 * i + 1] = -inner_product_matrix[inner_product_entry(3 * i + 2, gcounter)];
			}
		else
			for (i=0;i<tmp_length;i++)
				linear[i] = -inner_product_matrix[inner_product_entry(i, gcounter)];

		err = solve_system(system, linear, tmp_length, tmp_length, delta);
		if (sparse)
		{
			for (i=tmp_length / 2;i>0;i--)
			{
				delta[3 * i - 1] = delta[2 * i - 1];
				delta[3 * i - 2] = delta[2 * i - 2];
				delta[3 * i - 3] = 0;
			}
			tmp_length = 3 * DynamicM + 3;
		}
		origDirDeriv = 0;
		memset(dTdtmp, 0, sizeof(double) * Plength);
		dspmv_(U, &tmp_length, &one, inner_product_matrix, delta, &inc, &zero, dTdtmp, &inc);
		dTd = ddot_(&tmp_length, delta, &inc, dTdtmp, &inc);
		origDirDeriv = dTdtmp[gcounter];
		double ws = dTdtmp[wcounter];

		if (origDirDeriv > 0)
		{
			origDirDeriv = -origDirDeriv;
			ws = -ws;
			dscal_(&tmp_length, &minusone, delta, &inc);
		}
		if (fabs(origDirDeriv) < inner_product_matrix[inner_product_entry(gcounter, gcounter)] * 1e-12)
		{
			memset(delta, 0, sizeof(double) * Plength);
			delta[gcounter] = -1;
			dTd = inner_product_matrix[inner_product_entry(gcounter, gcounter)];
			origDirDeriv = -dTd;
			ws  = -inner_product_matrix[inner_product_entry(wcounter, gcounter)];
			memset(step, 0, sizeof(double) * length);
			memset(step_x, 0, sizeof(double) * l);
			daxpy_(&length, &minusone, g, &inc, step, &inc);
			daxpy_(&l, &minusone, gx, &inc, step_x, &inc);
		}
		else
		{
			dgemv_(N, &length, &tmp_length, &one, P, &length, delta, &inc, &zero, step, &inc);
			dgemv_(N, &l, &tmp_length, &one, Px, &l, delta, &inc, &zero, step_x, &inc);
		}

		int64_t iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		MPI_Allgatherv(step, length, MPI_DOUBLE, recv_buffer, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);
		fun_obj->Xv(recv_buffer, step_x);
		time_st = wall_clock_ns();

		double oldf = f;
		step_size = fun_obj->line_search(w, step, origDirDeriv, dTd, ws, &f, eta, 1.0, step_x);

		if (step_size == 0)
		{
			info("WARNING: step size too small\n");
			break;
		}
		double actred = oldf - f;
		if (fabs(actred) < 0)
		{
			info("WARNING: actred < 0\n");
			break;
		}
		if (fabs(actred) <= 1.0e-20*fabs(f))
		{
			info("WARNING: actred too small\n");
			break;
		}

		iter++;
		previous_M = DynamicM;
		DynamicM = min(iter, M);
		tmp_length = 3 * DynamicM + 3;

		wcounter = (wcounter + 3) % Plength;
		gcounter = wcounter + 1;
		dcounter = wcounter + 2;
		if (step_size == 1)
		{
			double *tmp_ptr = P + wcounter * length;
			for (i=0;i<length;i++)
				tmp_ptr[i] = localw[i] + step[i];
			localw = P + wcounter * length;
		}
		else
		{
			memcpy(P + wcounter * length, localw, sizeof(double) * length);
			localw = P + wcounter * length;
			daxpy_(&length, &step_size, step, &inc, localw, &inc);
		}
		wx = Px + wcounter * l;

		g = P + gcounter * length;
		gx = Px + gcounter * l;

		memcpy(w + start, localw, sizeof(double) * length);
		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		if (iter % 10 == 0 && iter != 0)
		{
			MPI_Allgatherv(localw, length, MPI_DOUBLE, recv_buffer, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);
			memcpy(w, recv_buffer, sizeof(double) * size_t(n));
			f = fun_obj->fun(w);
		}
		fun_obj->get_wx(wx);
		time_st = wall_clock_ns();
		fun_obj->grad(w, global_g);
		memcpy(g, global_g + start, sizeof(double) * size_t(length));

		D_inverse_g = P + dcounter * length;
		D_inverse_g_x = Px + dcounter * l;
		fun_obj->get_diagH(D);
		for (i=0;i<n;i++)
			D[i] = global_g[i] / D[i];
		fun_obj->Xv(global_g, gx, 2);
		memcpy(D_inverse_g, D + start, sizeof(double) * size_t(length));

		update_inner_products(inner_product_matrix, DynamicM, previous_M, P, delta, step_size);
	}
	MPI_Allgatherv(localw, length, MPI_DOUBLE, recv_buffer, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);
	memcpy(w, recv_buffer, sizeof(double) * size_t(n));

	delete[] dTdtmp;
	delete[] delta;
	delete[] system;
	delete[] Px;
	delete[] P;
	delete[] step;
	delete[] step_x;
	delete[] sqrt_C_times_D;
	delete[] linear;
	delete[] inner_product_matrix;
	delete[] sqrtD_Px;
	delete[] global_g;
}

void MLBFGS_DIAG::compute_system(double* inner_product_matrix, double* Px, double* sqrt_C_times_D, double* system, int sparse)
{
	double one = 1.0;
	double zero = 0;
	int i,j;
	char U[] = "U";
	char T[] = "T";
	double *tmp = new double[tmp_length * (tmp_length + 1) / 2];
	if (sparse)
		for (i=0;i<tmp_length/2;i++)
		{
			int counteri = 2 * i * l;
			int tmp_counter = (3 * i + 1) * l;
			for (j=0;j<l;j++)
				sqrtD_Px[counteri + j] = sqrt_C_times_D[j] * Px[tmp_counter + j];
			counteri += l;
			tmp_counter = (3 * i + 2) * l;
			for (j=0;j<l;j++)
				sqrtD_Px[counteri + j] = sqrt_C_times_D[j] * Px[tmp_counter + j];
		}
	else
		for (i=0;i<tmp_length;i++)
		{
			int counteri = i * l;
			for (j=0;j<l;j++)
				sqrtD_Px[counteri + j] = sqrt_C_times_D[j] * Px[counteri + j];
		}
	dsyrk_(U, T, &tmp_length, &l, &one, sqrtD_Px, &l, &zero, system, &tmp_length);

	int counter = 0;
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			tmp[counter++] = system[i * tmp_length + j];
	mpi_allreduce(tmp, counter, MPI_DOUBLE, MPI_SUM);
	counter = 0;
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[i * tmp_length + j] = tmp[counter++];

	if (sparse)
		for (i=0;i<tmp_length/2;i++)
		{
			for (j=0;j < i;j++)
			{
				system[2 * i * tmp_length + 2 * j] += inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 1)];
				system[2 * i * tmp_length + 2 * j + 1] += inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 2)];
			}
			system[2 * i * tmp_length + 2 * i] += inner_product_matrix[inner_product_entry(3 * i + 1, 3 * i + 1)];

			for (j=0;j <= i;j++)
			{
				system[(2 * i + 1) * tmp_length + 2 * j] += inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 1)];
				system[(2 * i + 1) * tmp_length + 2 * j + 1] += inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 2)];
			}
		}
	else
		for (i=0;i<tmp_length;i++)
			for (j=0;j<=i;j++)
				system[i * tmp_length + j] += inner_product_matrix[inner_product_entry(i,j)];

	for (i=0;i<tmp_length;i++)
		for (j=i+1;j<tmp_length;j++)
			system[i * tmp_length + j] = system[j * tmp_length + i];
	delete[] tmp;
}

void MLBFGS_DIAG::compute_system(double* inner_product_matrix, double* Px, double* sqrt_C_times_D, double* system, int sparse, int *index, int index_length)
{
	double *tmp = new double[tmp_length * (tmp_length + 1) / 2];
	double one = 1.0;
	double zero = 0;
	int i,j;
	char U[] = "U";
	char T[] = "T";
	if (sparse)
	{
		for (i=0;i<tmp_length/2;i++)
		{
			int counteri = 2 * i * index_length;
			int tmp_counter = (3 * i + 1) * l;
			for (j=0;j<index_length;j++)
				sqrtD_Px[counteri + j] = sqrt_C_times_D[j] * Px[tmp_counter + index[j]];
			counteri += index_length;
			tmp_counter += l;
			for (j=0;j<index_length;j++)
				sqrtD_Px[counteri + j] = sqrt_C_times_D[j] * Px[tmp_counter + index[j]];
		}
	}
	else
		for (i=0;i<tmp_length;i++)
		{
			int counteri = i * index_length;
			int tmp_counter = i * l;
			for (j=0;j<index_length;j++)
				sqrtD_Px[counteri + j] = sqrt_C_times_D[j] * Px[tmp_counter + index[j]];
		}

	dsyrk_(U, T, &tmp_length, &index_length, &one, sqrtD_Px, &index_length, &zero, system, &tmp_length);

	int counter = 0;
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			tmp[counter++] = system[i * tmp_length + j];
	mpi_allreduce(tmp, counter, MPI_DOUBLE, MPI_SUM);
	counter = 0;
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[i * tmp_length + j] = tmp[counter++];

	if (sparse)
		for (i=0;i<tmp_length/2;i++)
		{
			for (j=0;j < i;j++)
			{
				system[2 * i * tmp_length + 2 * j] += inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 1)];
				system[2 * i * tmp_length + 2 * j + 1] += inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 2)];
			}
			system[2 * i * tmp_length + 2 * i] += inner_product_matrix[inner_product_entry(3 * i + 1, 3 * i + 1)];

			for (j=0;j <= i;j++)
			{
				system[(2 * i + 1) * tmp_length + 2 * j] += inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 1)];
				system[(2 * i + 1) * tmp_length + 2 * j + 1] += inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 2)];
			}
		}
	else
		for (i=0;i<tmp_length;i++)
			for (j=0;j<=i;j++)
				system[i * tmp_length + j] += inner_product_matrix[inner_product_entry(i,j)];

	for (i=0;i<tmp_length;i++)
		for (j=i+1;j<tmp_length;j++)
			system[i * tmp_length + j] = system[j * tmp_length + i];
	delete[] tmp;
}


void MLBFGS_DIAG::update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* delta, double step_size)
{
	int i;
	int prev_length = 3 * previous_M + 3;
	int inc = 1;
	double *PTg = new double[tmp_length * 2];
	double *PTd = PTg + tmp_length;
	double *PTw = new double[tmp_length];
	double one = 1.0;
	double zero = 0;
	char U[] = "U";
	char T[] = "T";
	dgemv_(T, &length, &tmp_length, &one, P, &length, P + gcounter * length, &inc, &zero, PTg, &inc);
	dgemv_(T, &length, &tmp_length, &one, P, &length, P + dcounter * length, &inc, &zero, PTd, &inc);
	mpi_allreduce(PTg, tmp_length * 2, MPI_DOUBLE, MPI_SUM);


	int prev_w = wcounter - 3;
	if (wcounter == 0)
		prev_w = 3 * DynamicM;

	if (step_size != 1)
		dscal_(&prev_length, &step_size, delta, &inc);
	delta[prev_w] += 1;

	dspmv_(U, &prev_length, &one, inner_product_matrix, delta, &inc, &zero, PTw, &inc);
	PTw[wcounter] = ddot_(&prev_length, PTw, &inc, delta, &inc);

	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, wcounter)] = PTw[i];
	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, gcounter)] = PTg[i];
	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, dcounter)] = PTd[i];

	delete[] PTw;
	delete[] PTg;
}

LBFGS::LBFGS(const function *fun_obj, double eps, int m, double eta, int max_iter): SOLVER(fun_obj, eps, max_iter)
{
	this->eta = eta;
	this->M = m;
	offset = 0;
}

LBFGS::~LBFGS()
{
}

void LBFGS::solve(double *w)
{
	int nr_node = mpi_get_size();
	int n = fun_obj->get_nr_variable();
	int shift = int(ceil(double(n) / double(nr_node)));
	int i;
	int k = 0;
	double f;
	int search = 1, iter = 1, inc = 1;
	double minusone = -1;
	double *g;
	double *old_g;
	double init_size = 1;
	double *step = new double[n];
	double *local_step = new double[length];
	memset(w, 0, sizeof(double) * size_t(n));
	int64_t time_st = wall_clock_ns();
	f = fun_obj->fun(w);
	g = new double[n];
	fun_obj->grad(w, g);
	double gnorm1 = dnrm2_(&n, g, &inc);
	double gnorm = gnorm1;

	if (gnorm <= eps*gnorm1)
		search = 0;

	iter = 0;

	double **s = new double*[M];
	double **y = new double*[M];
	double **inner_product_matrix = new double*[2*M+1];
	double *rho = new double[M];
	double* delta = new double[M * 2 + 1];
	int *recv_count = new int[nr_node];
	int *displace = new int[nr_node];
	int counter = 0;
	for (i=0;i<nr_node;i++)
	{
		recv_count[i] = shift;
		displace[i] = counter;
		counter += shift;
		if (counter >= n)
		{
			counter = 0;
			shift = 0;
		}
		else if (counter + shift > n)
			shift = n - counter;
	}
	double *recv_buffer = new double[n];
	for (int i = 0; i < M; i++)
	{
		y[i] = new double[length];
		s[i] = new double[length];
	}
	for (int i=0; i < 2 * M + 1; i++)
		inner_product_matrix[i] = new double[2*M+1];

	double origDirDeriv = 0.0;
	double one = 1.0;
	double step_size = 0.0;
	double dTd = 0;
	int previous_M = 0;
	int64_t iter_ed = wall_clock_ns();
	double tt = wall_time_diff(iter_ed, time_st);
	info("iter %d m %d |g| %5.3e f %15.20e step %5.3e time %5.3e\n", iter, 0, gnorm, f, 0.0, tt);
	while (iter < max_iter && search)
	{
		int DynamicM = min(iter, M);
		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		time_st = wall_clock_ns();

		if (iter > 0)
		{
			g = new double[n];
			fun_obj->grad(w, g);
			double s0y0 = 0;

			memcpy(y[k], g + start, sizeof(double) * size_t(length));
			daxpy_(&length, &minusone, old_g + start, &inc, y[k], &inc);

			delete[] old_g;

			update_inner_products(inner_product_matrix, k, n, DynamicM, previous_M, s, y, g, delta, step_size);
			gnorm = sqrt(inner_product_matrix[2 * DynamicM][2 * DynamicM]);
			s0y0 = inner_product_matrix[2 * k][2* k + 1];
			info("iter %d m %d |g| %5.3e f %15.20e step %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, tt);
			if (gnorm <= eps*gnorm1)
				break;
			if (s0y0 == 0)
				break;
			rho[k] = 1.0 / s0y0;
			k = (k+1)%M;

			TwoLoop(inner_product_matrix, rho, DynamicM, k, n, delta);
			local_step = s[k];
			if (iter >= M)
			{
				dscal_(&length, delta + (2 * k), local_step, &inc);
				daxpy_(&length, delta + (2 * k + 1), y[k], &inc, local_step, &inc);
			}
			else
				memset(local_step, 0, sizeof(double) * size_t(length));
			daxpy_(&length, delta + (2 * DynamicM), g + start, &inc, local_step, &inc);
			for (i=0;i<k;i++)
			{
				daxpy_(&length, delta + (2 * i), s[i], &inc, local_step, &inc);
				daxpy_(&length, delta + (2 * i + 1), y[i], &inc, local_step, &inc);
			}
			for (i=k+1;i<DynamicM;i++)
			{
				daxpy_(&length, delta + (2 * i), s[i], &inc, local_step, &inc);
				daxpy_(&length, delta + (2 * i + 1), y[i], &inc, local_step, &inc);
			}
			MPI_Allgatherv(local_step, length, MPI_DOUBLE, step, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);

			origDirDeriv = 0;
			dTd = 0;
			int tmp_length = 2 * DynamicM + 1;
			origDirDeriv = ddot_(&tmp_length, delta, &inc, inner_product_matrix[2 * DynamicM], &inc);
			for (i=0;i<2 * DynamicM + 1;i++)
				dTd += delta[i] * ddot_(&tmp_length, delta, &inc, inner_product_matrix[i], &inc);
			init_size = 1;
		}
		else
		{
			local_step = s[0];
			memcpy(local_step, g + start, sizeof(double) * size_t(length));
			dscal_(&length, &minusone, local_step, &inc);
			memcpy(step, g, sizeof(double) * size_t(n));
			dscal_(&n, &minusone, step, &inc);
			init_size = 1.0 / gnorm;
			dTd = gnorm * gnorm;
			origDirDeriv = -gnorm * gnorm;
		}

		if (origDirDeriv >= 0)
		{
			info("WARNING: not a descent direction\n");
			break;
		}
		double ws = ddot_(&length, w + start, &inc, local_step, &inc);
		mpi_allreduce(&ws, 1, MPI_DOUBLE, MPI_SUM);

		step_size = fun_obj->line_search(w, step, origDirDeriv, dTd, ws, &f, eta, init_size);
		if (iter == 0)
		{
			inner_product_matrix[0][0] = dTd;
			delta[0] = -1;
		}
		if (step_size == 0)
		{
			info("WARNING: step size too small\n");
			break;
		}
		dscal_(&length, &step_size, local_step, &inc);
		daxpy_(&length, &one, local_step, &inc, w + start, &inc);
		old_g = g;
		iter++;
		previous_M = DynamicM;
	}

	MPI_Allgatherv(w + start, length, MPI_DOUBLE, recv_buffer, recv_count, displace, MPI_DOUBLE, MPI_COMM_WORLD);
	memcpy(w, recv_buffer, sizeof(double) * size_t(n));

	delete[] g;
	delete[] recv_count;
	delete[] recv_buffer;
	delete[] displace;
	delete[] step;
	for (int i = 0; i < M; i++)
	{
		delete[] y[i];
		delete[] s[i];
	}
	delete[] y;
	delete[] s;
	for (int i=0; i < 2 * M + 1; i++)
		delete[] inner_product_matrix[i];
	delete[] inner_product_matrix;
	delete[] rho;
	delete[] delta;

}

void LBFGS::update_inner_products(double** inner_product_matrix, int k, int n, int DynamicM, int previous_M, double** s, double** y, double* g, double* delta, double step_size)
{
	int i;
	int inc = 1;
	int prev_length = 2 * previous_M + 1;
	double *tmp = new double[prev_length + 1];
	double *tmp2 = new double[2 * DynamicM];

	double* buffer = new double[DynamicM * 2 + 2];
	memset(buffer, 0, sizeof(double) * size_t(DynamicM * 2 + 2));
	for (i=0;i<prev_length;i++)
		tmp[i] = ddot_(&prev_length, delta, &inc, inner_product_matrix[i], &inc);
	tmp[prev_length] = ddot_(&prev_length, delta, &inc, tmp, &inc) * step_size * step_size;
	for (i=0;i<prev_length;i++)
		tmp[i] *= step_size;

	for (i=0;i<DynamicM;i++)
	{
		buffer[2 * i] = ddot_(&length, g + start, &inc, s[i] + offset, &inc);
		buffer[2 * i + 1] = ddot_(&length, g + start, &inc, y[i] + offset, &inc);
	}
	buffer[2 * DynamicM] = ddot_(&length, g + start, &inc, g + start, &inc);
	buffer[2 * DynamicM + 1] = ddot_(&length, y[k] + offset, &inc, y[k] + offset, &inc);
	mpi_allreduce(buffer, 2 * DynamicM + 2, MPI_DOUBLE, MPI_SUM);
	for (i=0;i<prev_length - 1; i++)
		tmp2[i] = buffer[i] - inner_product_matrix[prev_length-1][i];

	tmp2[2 * k] = buffer[2 * k] - tmp[prev_length - 1];
//	s[k] * y[k] = s[k] * g - s[k] * old_g

	tmp2[2 * k + 1] = buffer[2 * DynamicM + 1];
//	y[k] * y[k]

	for (i=0;i< prev_length - 1;i++)
	{
		inner_product_matrix[2 * k][i] = tmp[i];
		inner_product_matrix[i][2 * k] = tmp[i];
	}

	for (i=0;i< 2 * DynamicM;i++)
	{
		inner_product_matrix[2 * k + 1][i] = tmp2[i];
		inner_product_matrix[i][2 * k + 1] = tmp2[i];
	}

	inner_product_matrix[2 * k][2 * k] = tmp[prev_length];

	for (i=0; i< 2 * DynamicM + 1; i++)
	{
		inner_product_matrix[2 * DynamicM][i] = buffer[i];
		inner_product_matrix[i][2 * DynamicM] = buffer[i];
	}

	delete[] buffer;
	delete[] tmp;
	delete[] tmp2;
}

void LBFGS::TwoLoop(double** inner_product_matrix, double* rho, int DynamicM, int k, int n, double *delta)
{
	int i, j;
	int inc = 1;
	int start = k-1;
	double beta;
	double *alpha = new double[DynamicM];
	int tmp_length = 2 * DynamicM + 1;

	memset(delta, 0, sizeof(double) * size_t(2 * DynamicM));
	memset(alpha, 0, sizeof(double) * size_t(DynamicM));
	delta[2 * DynamicM] = -1;
	if (k < DynamicM)
		start += DynamicM;
	int lastrho = start % DynamicM;

	for (i = 0; i < DynamicM; i++)
	{
		j = start % DynamicM;

		start--;
		if (rho[j] > 0)
		{
			alpha[j] = rho[j] * ddot_(&tmp_length, delta, &inc, inner_product_matrix[2 * j], &inc);
			delta[2 * j + 1] -= alpha[j];
		}
		else
		{
			fprintf(stderr,"ERROR: rho[%d] <= 0\n",i);
			return;
		}
	}
	if (lastrho != -1)
	{
		double scal = (1.0 / (rho[lastrho] * inner_product_matrix[2 * lastrho + 1][2 * lastrho + 1]));
		dscal_(&tmp_length, &scal, delta, &inc);
		for (i = 0; i < DynamicM; i++)
		{
			start++;
			j = start % DynamicM;
			beta = alpha[j] - rho[j] * ddot_(&tmp_length, delta, &inc, inner_product_matrix[2 * j + 1], &inc);
			delta[2 * j] += beta;
		}
	}
	delete[] alpha;
}

TRON::TRON(const function *fun_obj, double eps, double default_eps, double eps_cg, int max_iter): SOLVER(fun_obj, eps, max_iter)
{
	this->eps_cg=eps_cg;
	this->default_eps = default_eps;
}

TRON::~TRON()
{
}

void TRON::solve(double *w)
{
	// Parameters for updating the iterates.
	double eta0 = 1e-4, eta1 = 0.25, eta2 = 0.75;

	// Parameters for updating the trust region size delta.
	double sigma1 = 0.25, sigma2 = 0.5, sigma3 = 4;

	int n = fun_obj->get_nr_variable();
	int i, cg_iter=0;
	double delta=0, sMnorm, one=1.0;
	int printed = 0;
	double alpha, f, fnew, prered = 0, actred = 0, gs;
	int search = 1, iter = 1, inc = 1;
	double *s = new double[n];
	double *r = new double[n];
	double *g = new double[n];

	const double alpha_pcg = 0.01;
	double *M = new double[n];

	// calculate gradient norm at w=0 for stopping condition.
	memset(w, 0, sizeof(double) * size_t(n));

	int64_t time_st = wall_clock_ns();
	f = fun_obj->fun(w);
	fun_obj->grad(w, g);
	delta = dnrm2_(&n, g, &inc);
	double gnorm = delta;
	double gnorm0 = gnorm;

	if (gnorm <= eps*gnorm0)
		search = 0;

	iter = 0;
	int64_t iter_ed = wall_clock_ns();
	double tt = wall_time_diff(iter_ed, time_st);
	info("iter %2d act %5.3e pre %5.3e delta %5.3e f %15.20e |g| %5.3e CG %3d time %5.3e\n", iter, 0.0, 0.0, delta, f, gnorm, 0, tt);
	time_st = wall_clock_ns();

	double *w_new = new double[n];
	bool reach_boundary;
	double size = (double) mpi_get_size();
	while (iter <= max_iter && search)
	{
		fun_obj->get_diagH(M);
		for(i=0; i<n; i++)
			M[i] = (1-alpha_pcg) + alpha_pcg*M[i];
		if (iter == 1)
			delta = sqrt(uTMv(n, g, M, g));
		mpi_allreduce(&delta, 1, MPI_DOUBLE, MPI_SUM);
		delta /= size;
		cg_iter = trpcg(delta, g, M, s, r, &reach_boundary);

		memcpy(w_new, w, sizeof(double) * size_t(n));
		daxpy_(&n, &one, s, &inc, w_new, &inc);

		double buffer[2];
		gs = ddot_(&length, g + start, &inc, s + start, &inc);
		buffer[0] = gs;
		buffer[1] = ddot_(&length, s + start, &inc, r + start, &inc);
		mpi_allreduce(buffer, 2, MPI_DOUBLE, MPI_SUM);
		gs = buffer[0];
		prered = -0.5*(gs-buffer[1]);
		fnew = fun_obj->fun(w_new);

		// Compute the actual reduction.
		actred = f - fnew;

		// On the first iteration, adjust the initial step bound.
		sMnorm = sqrt(uTMv(n, s, M, s));
		if (iter == 1)
			delta = min(delta, sMnorm);

		// Compute prediction alpha*snorm of the step.
		if (fnew - f - gs <= 0)
			alpha = sigma3;
		else
			alpha = max(sigma1, -0.5*(gs/(fnew - f - gs)));

		// Update the trust region bound according to the ratio of actual to predicted reduction.
		if (actred < eta0*prered)
			delta = min(alpha*sMnorm, sigma2*delta);
		else if (actred < eta1*prered)
			delta = max(sigma1*delta, min(alpha*sMnorm, sigma2*delta));
		else if (actred < eta2*prered)
			delta = max(sigma1*delta, min(alpha*sMnorm, sigma3*delta));
		else
		{
			if (reach_boundary)
				delta = sigma3*delta;
			else
				delta = max(delta, min(alpha*sMnorm, sigma3*delta));
		}

		if (actred > eta0*prered)
		{
			iter++;
			memcpy(w, w_new, sizeof(double) * size_t(n));
			f = fnew;
			fun_obj->grad(w, g);

			gnorm = dnrm2_(&n, g, &inc);
			iter_ed = wall_clock_ns();
			tt += wall_time_diff(iter_ed, time_st);
			info("iter %2d act %5.3e pre %5.3e delta %5.3e f %15.20e |g| %5.3e CG %3d time %5.3e\n", iter, actred, prered, delta, f, gnorm, cg_iter, tt);
			time_st = wall_clock_ns();
			if (gnorm <= default_eps * gnorm0 && printed == 0)
			{
				printed = 1;
				info("Default stopping value = %15.20e\n",f);
			}
			if (gnorm <= default_eps * 0.1 * gnorm0 && printed < 2)
			{
				printed++;
				info("10 times tighter stopping value = %15.20e\n",f);
			}
			if (gnorm <= default_eps * 0.01 * gnorm0 && printed < 3)
			{
				printed++;
				info("100 times tighter stopping value = %15.20e\n",f);
			}


			if (gnorm <= eps*gnorm0)
				break;
		}
		if (f < -1.0e+32)
		{
			info("WARNING: f < -1.0e+32\n");
			break;
		}
		if (prered <= 0)
		{
			info("WARNING: prered <= 0\n");
			break;
		}
		if (fabs(actred) <= 1.0e-12*fabs(f) &&
			fabs(prered) <= 1.0e-12*fabs(f))
		{
			info("WARNING: actred and prered too small\n");
			break;
		}
	}
	info("iter %2d act %5.3e pre %5.3e delta %5.3e f %15.20e |g| %5.3e CG %3d time %5.3e\n", iter, actred, prered, delta, f, gnorm, cg_iter, tt);
	if (gnorm <= default_eps * gnorm0 && printed == 0)
	{
		printed = 1;
		info("Default stopping value = %15.20e\n",f);
	}

	delete[] g;
	delete[] r;
	delete[] w_new;
	delete[] s;
}

int TRON::trpcg(double delta, double *g, double *M, double *s, double *r, bool *reach_boundary)
{
	int i, inc = 1;
	int n = fun_obj->get_nr_variable();
	double one = 1;
	double *d = new double[n];
	double *Hd = new double[n];
	double zTr, znewTrnew, alpha, beta, cgtol;
	double *z = new double[n];
	double size = (double) mpi_get_size();

	*reach_boundary = false;
	for (i=0; i<n; i++)
	{
		s[i] = 0;
		r[i] = -g[i];
		z[i] = r[i] / M[i];
		d[i] = z[i];
	}

	zTr = ddot_(&n, z, &inc, r, &inc);
	mpi_allreduce(&zTr, 1, MPI_DOUBLE, MPI_SUM);
	zTr /= size;
	cgtol = eps_cg*sqrt(zTr);
	int cg_iter = 0;

	while (1)
	{
		if (sqrt(zTr) <= cgtol)
			break;
		cg_iter++;
		fun_obj->Hv(d, Hd);

		alpha = zTr/ddot_(&n, d, &inc, Hd, &inc);
		mpi_allreduce(&alpha, 1, MPI_DOUBLE, MPI_SUM);
		alpha /= size;
		daxpy_(&n, &alpha, d, &inc, s, &inc);

		double sMnorm = sqrt(uTMv(n, s, M, s));
		mpi_allreduce(&sMnorm, 1, MPI_DOUBLE, MPI_SUM);
		sMnorm /= size;
		if (sMnorm > delta)
		{
			info("cg reaches trust region boundary\n");
			*reach_boundary = true;
			alpha = -alpha;
			daxpy_(&n, &alpha, d, &inc, s, &inc);

			double sTMd = uTMv(n, s, M, d);
			double sTMs = uTMv(n, s, M, s);
			double dTMd = uTMv(n, d, M, d);
			double dsq = delta*delta;
			double rad = sqrt(sTMd*sTMd + dTMd*(dsq-sTMs));
			if (sTMd >= 0)
				alpha = (dsq - sTMs)/(sTMd + rad);
			else
				alpha = (rad - sTMd)/dTMd;
			mpi_allreduce(&alpha, 1, MPI_DOUBLE, MPI_SUM);
			alpha /= size;
			daxpy_(&n, &alpha, d, &inc, s, &inc);
			alpha = -alpha;
			daxpy_(&n, &alpha, Hd, &inc, r, &inc);
			break;
		}
		alpha = -alpha;
		daxpy_(&n, &alpha, Hd, &inc, r, &inc);

		for (i=0; i<n; i++)
			z[i] = r[i] / M[i];
		znewTrnew = ddot_(&n, z, &inc, r, &inc);
		mpi_allreduce(&znewTrnew, 1, MPI_DOUBLE, MPI_SUM);
		znewTrnew /= size;
		beta = znewTrnew/zTr;
		dscal_(&n, &beta, d, &inc);
		daxpy_(&n, &one, z, &inc, d, &inc);
		zTr = znewTrnew;
	}

	delete[] d;
	delete[] Hd;
	delete[] z;

	return(cg_iter);
}


double TRON::norm_inf(int n, double *x)
{
	double dmax = fabs(x[0]);
	for (int i=1; i<n; i++)
		if (fabs(x[i]) >= dmax)
			dmax = fabs(x[i]);
	return(dmax);
}
