#ifndef _SOLVER_H
#define _SOLVER_H

class function
{
public:
	virtual double fun(double *w) = 0 ; //Evaluate f(w)
	virtual void grad(double *w, double *g) = 0 ;
	virtual void Hv(double *s, double *Hs) = 0 ;
	virtual void Xv(double *v, double *Xv) = 0; //Xv = X*v
	virtual void Xv(double *v, double *Xv, int vecs) = 0; //Xv with v being a vector of n * vecs. Used in LCOMM_DIAG
	virtual void sqrt_Hessian_coeff(double *sqrt_C_times_D, int **index, int *index_length) = 0; //Compute the the square root of the diagonal matrix D_w in Sec 5.2 of the paper, stored in vector form
	virtual double line_search(double *p, double gTp, double pTp, double wTp, double *f, double eta, double init_size = 1.0, double *Xp = NULL) = 0 ;// w is the current iterate, p the update direction, g the gradient at w, Xd the vector storing X^T d; on return, fnew is the new objective value and the return value is the step size, but w is not updated

	virtual int get_nr_variable(void) = 0 ;
	virtual void get_wx(double *external_wx) = 0 ;
	virtual void get_diagH(double *M) = 0 ;
	virtual ~function(void){}
};

class SOLVER //Abstract class for solvers
{
public:
	SOLVER(const function *fun_obj, double eps = 0.1, int max_iter = 500000);// eps for the stopping condition and max_iter is the max number of iterations allowed
	virtual void solve(double *w) = 0 ;//Solve the optimization problem
	void set_print_string(void (*i_print) (const char *buf));//Either output messages to stdout or don't output any message
	virtual ~SOLVER(void){}
protected:
	void info(const char *fmt,...);//Function for flush output to stdout
	double eps;
	int max_iter;
	function *fun_obj;//objective function handle
	void (*solver_print_string)(const char *buf);
};

class LCOMM_BFGS: public SOLVER
{
public:
	LCOMM_BFGS(const function *fun_obj, int l, double eps = 0.1, int m=10, double eta = 1e-2, int threads  = 1, int max_iter = 500000);
	~LCOMM_BFGS();

	void solve(double *w);
protected:
	int M;
	int inner_product_entry(int i, int j);
	double eta;
	int l;
	int n;
	int Plength;
	int tmp_length;
	int wcounter;
	int gcounter;

	//For parallelizing operations on vectors of size n and l, respectively
	int num_blocks_n;
	int *block_size_n;
	int *block_start_n;

	int num_blocks_l;
	int *block_size_l;
	int *block_start_l;

	double *sqrtD_XTP;
	double **tmp_sys;//For parallelization in linear system construction

private:
	void update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* t, double step_size);// Update the matrix P^T A P stored in inner_product_matrix. t is the coefficients for forming the update direction (minimizer of (3)). DynamicM is the current number of historical iterations used (DynamicM = min(M, iter)). previous_M is the value of DynamicM in last iteration

	void compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse);// Compute the coefficients in the linear system; double *system stores the output. sparse = 1 if DynamicM < M, and in this case the columns storing w_i, i > 0 can be ignored in the linear system as they are linear combinations of other columns
	void compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse, int *index, int index_length);// Compute the coefficients in the linear system when D_w is sparse, with the index set given in *index and index_length is the size of the index set. sparse = 1 if DynamicM < M, and in this case the columns storing w_i, i > 0 can be ignored in the linear system as they are linear combinations of other columns
};

class LCOMM_DIAG: public LCOMM_BFGS
{
public:
	LCOMM_DIAG(const function *fun_obj, int l, double eps = 0.1, int m=10, double eta = 1e-2, int threads = 1, int max_iter = 500000);
	~LCOMM_DIAG();

	void solve(double *w);
private:
	int dcounter;//The column in P that maintains the current D^{-1} grad f
	void update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* t, double step_size);// Update the matrix P^T A P stored in inner_product_matrix. t is the coefficients for forming the update direction (minimizer of (3)). DynamicM is the current number of historical iterations used (DynamicM = min(M, iter)). previous_M is the value of DynamicM in last iteration
	void compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse);// Compute the coefficients in the linear system; double *system stores the output. sparse = 1 if DynamicM < M, and in this case the columns storing w_i, i > 0 can be ignored in the linear system as they are linear combinations of other columns
	void compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse, int *index, int index_length);// Compute the coefficients in the linear system when D_w is sparse, with the index set given in *index and index_length is the size of the index set. sparse = 1 if DynamicM < M, and in this case the columns storing w_i, i > 0 can be ignored in the linear system as they are linear combinations of other columns
};

class TRON: public SOLVER
{
public:
	TRON(const function *fun_obj, double eps = 0.1, double default_eps = 0.01, double eps_cg = 0.1, int max_iter = 10000, int max_total_cg_iter = 500000);
	~TRON();

	void solve(double *w);

private:
	int trpcg(double delta, double *g, double *M, double *s, double *r, bool *reach_boundary);
	double norm_inf(int n, double *x);

	int max_total_cg_iter;
	double eps_cg;
	double default_eps;
};

#endif
